@@ -23,11 +23,12 @@ import logging
 from typing import Dict, List
 import math
 from collections import defaultdict
+import warnings
 
+import numpy as np
 import xgboost as xgb
 import ray
-from ray.services import get_node_ip_address
-from ray.util.placement_group import placement_group, remove_placement_group
+from ray.util import get_node_ip_address
 import pandas
 
 from modin.distributed.dataframe.pandas import from_partitions
@@ -36,10 +37,10 @@ from .utils import RabitContext, RabitContextManager
 LOGGER = logging.getLogger("[modin.xgboost]")
 
 
-@ray.remote
+@ray.remote(num_cpus=0)
 class ModinXGBoostActor:
     """
-    Ray actor-class runs training/prediction on remote workers.
+    Ray actor-class runs training on the remote worker.
 
     Parameters
     ----------
@@ -88,17 +89,6 @@ class ModinXGBoostActor:
 
         return xgb.DMatrix(X, y, nthread=self._nthreads)
 
-    def get_ip(self):
-        """
-        Get IP address of actor.
-
-        Returns
-        -------
-        str
-            An IP address of node used by actor.
-        """
-        return get_node_ip_address()
-
     def set_train_data(self, *X_y, add_as_eval_method=None):
         """
         Set train data for actor.
@@ -117,25 +107,6 @@ class ModinXGBoostActor:
         if add_as_eval_method is not None:
             self._evals.append((self._dtrain, add_as_eval_method))
 
-    def set_predict_data(
-        self,
-        *X,
-    ):
-        """
-        Set prediction data for actor.
-
-        Parameters
-        ----------
-        *X : iterable
-            Sequence of ray.ObjectRef objects. When it is passed in actor, auto-materialization
-            of ray.ObjectRef -> pandas.DataFrame happens.
-        """
-        X = pandas.concat(X, axis=0)
-        self._dpredict = {
-            "dmatrix": xgb.DMatrix(X, nthread=self._nthreads),
-            "index": X.index,
-        }
-
     def add_eval_data(self, *X_y, eval_method):
         """
         Add evaluation data for actor.
@@ -197,36 +168,6 @@ class ModinXGBoostActor:
             LOGGER.info(f"Local training time: {time.time() - s} s")
             return {"booster": bst, "history": evals_result}
 
-    def predict(self, booster: xgb.Booster, **kwargs):
-        """
-        Run local XGBoost prediction.
-
-        Parameters
-        ----------
-        booster : xgboost.Booster
-            A trained booster.
-        **kwargs : dict
-            Other parameters for `xgboost.Booster.predict`.
-
-        Returns
-        -------
-        tuple
-            Pair of IP address of caller and pandas.DataFrame
-            with partial prediction result.
-        """
-        local_dpredict = self._dpredict
-        booster.set_param({"nthread": self._nthreads})
-
-        s = time.time()
-
-        predictions = pandas.DataFrame(
-            booster.predict(local_dpredict["dmatrix"], **kwargs),
-            index=local_dpredict["index"],
-        )
-        LOGGER.info(f"Local prediction time: {time.time() - s} s")
-
-        return get_node_ip_address(), predictions
-
 
 def _get_cluster_cpus():
     """
@@ -277,18 +218,15 @@ def _get_cpus_per_actor(num_actors):
     return cpus_per_actor
 
 
-def _get_num_actors(num_actors):
+def _get_num_actors(num_actors=None):
     """
-    Get number on actors to create.
-
-    In case `num_actors` is 'default_train' string,
-    integer number of actors will be computed  by condition
-    2 CPUs per 1 actor, in case 'default_predict' 8 CPUs per 1 actor.
+    Get number of actors to create.
 
     Parameters
     ----------
-    num_actors : int or {'default_train', 'default_predict'}
-        Number of actors to create or str to set defaults.
+    num_actors : int, optional
+        Desired number of actors. If is None, integer number of actors
+        will be computed by condition 2 CPUs per 1 actor.
 
     Returns
     -------
@@ -296,47 +234,18 @@ def _get_num_actors(num_actors):
         Number of actors to create.
     """
     min_cpus_per_node = _get_min_cpus_per_node()
-    if num_actors == "default_train":
+    if num_actors is None:
         num_actors_per_node = max(1, int(min_cpus_per_node // 2))
         return num_actors_per_node * len(ray.nodes())
-    elif num_actors == "default_predict":
-        num_actors_per_node = max(1, int(min_cpus_per_node // 8))
-        return num_actors_per_node * len(ray.nodes())
     elif isinstance(num_actors, int):
+        assert (
+            num_actors % len(ray.nodes()) == 0
+        ), "`num_actors` must be a multiple to number of nodes in Ray cluster."
         return num_actors
     else:
         RuntimeError("`num_actors` must be int or None")
 
 
-def _create_placement_group(num_cpus_per_actor, num_actors):
-    """
-    Create Ray placement group to grab resources.
-
-    Parameters
-    ----------
-    num_cpus_per_actor : int
-        Number of CPUs per actor.
-    num_actors : int
-        Number of actors.
-
-    Returns
-    -------
-    ray.util.PlacementGroup
-        Placement group with grabbed resources.
-    """
-    cpu_bundle = {"CPU": num_cpus_per_actor}
-    bundles = [cpu_bundle for _ in range(num_actors)]
-
-    pg = placement_group(bundles, strategy="SPREAD")
-
-    ready, _ = ray.wait([pg.ready()], timeout=100)
-
-    if ready is None:
-        raise TimeoutError("Placement group creation timeout.")
-
-    return pg
-
-
 def create_actors(num_actors):
     """
     Create ModinXGBoostActors.
@@ -348,22 +257,33 @@ def create_actors(num_actors):
 
     Returns
     -------
-    tuple
-        Pair of actors list and placement group of actors.
+    list
+        List of pairs (ip, actor).
     """
     num_cpus_per_actor = _get_cpus_per_actor(num_actors)
-    pg = _create_placement_group(num_cpus_per_actor, num_actors)
-    actors = [
-        ModinXGBoostActor.options(
-            num_cpus=num_cpus_per_actor, placement_group=pg
-        ).remote(i, nthread=num_cpus_per_actor)
-        for i in range(num_actors)
+    node_ips = [
+        key for key in ray.cluster_resources().keys() if key.startswith("node:")
     ]
-    return actors, pg
+    num_actors_per_node = num_actors // len(node_ips)
+    actors_ips = [ip for ip in node_ips for _ in range(num_actors_per_node)]
+
+    actors = [
+        (
+            node_ip.split("node:")[-1],
+            ModinXGBoostActor.options(resources={node_ip: 0.01}).remote(
+                i, nthread=num_cpus_per_actor
+            ),
+        )
+        for i, node_ip in enumerate(actors_ips)
+    ]
+    return actors
 
 
 def _split_data_across_actors(
-    actors: List, set_func, X_parts, y_parts=None, is_predict=False
+    actors: List,
+    set_func,
+    X_parts,
+    y_parts,
 ):
     """
     Split row partitions of data between actors.
@@ -376,49 +296,35 @@ def _split_data_across_actors(
         The function for setting data in actor.
     X_parts : list
         Row partitions of X data.
-    y_parts : list, optional
+    y_parts : list
         Row partitions of y data.
-    is_predict : bool, default: False
-        Is split data for predict or not.
     """
     X_parts_by_actors = _assign_row_partitions_to_actors(
         actors,
         X_parts,
-        is_predict=is_predict,
     )
 
-    if y_parts is not None:
-        y_parts_by_actors = _assign_row_partitions_to_actors(
-            actors,
-            y_parts,
-            data_for_aligning=X_parts_by_actors,
-        )
+    y_parts_by_actors = _assign_row_partitions_to_actors(
+        actors,
+        y_parts,
+        data_for_aligning=X_parts_by_actors,
+    )
 
-    for rank, actor in enumerate(actors):
-        X_parts = X_parts_by_actors[rank] if is_predict else X_parts_by_actors[rank][0]
-        if y_parts is None:
-            set_func(actor, *X_parts)
-        else:
-            y_parts = y_parts_by_actors[rank][0]
-            set_func(actor, *(X_parts + y_parts))
+    for rank, (_, actor) in enumerate(actors):
+        set_func(actor, *(X_parts_by_actors[rank][0] + y_parts_by_actors[rank][0]))
 
 
 def _assign_row_partitions_to_actors(
     actors: List,
     row_partitions,
     data_for_aligning=None,
-    is_predict=False,
 ):
     """
     Assign row_partitions to actors.
 
-    In case of `is_predict` == False, `row_partitions` will be
-    assigned to actors according to their IPs. If distribution isn't
-    even, partitions will be moved from actor with excess partitions
-    to actor with lack of them.
-
-    In case of `is_predict` == True, `row_partitions` will be
-    assigned evenly to actors in order.
+    `row_partitions` will be assigned to actors according to their IPs.
+    If distribution isn't even, partitions will be moved from actor
+    with excess partitions to actor with lack of them.
 
     Parameters
     ----------
@@ -429,8 +335,6 @@ def _assign_row_partitions_to_actors(
     data_for_aligning : dict, optional
         Data according to the order of which should be
         distributed `row_partitions`. Used to align y with X.
-    is_predict : bool, default: False
-        Is split data for predict or not.
 
     Returns
     -------
@@ -439,141 +343,115 @@ def _assign_row_partitions_to_actors(
         as {actor_rank: (partitions, order)}.
     """
     num_actors = len(actors)
-    if not is_predict:
-        if data_for_aligning is None:
-            parts_ips_ref, parts_ref = zip(*row_partitions)
+    if data_for_aligning is None:
+        parts_ips_ref, parts_ref = zip(*row_partitions)
 
-            # Group actors which are one the same ip
-            actor_ips = defaultdict(list)
+        # Group actors which are one the same ip
+        actor_ips = defaultdict(list)
+        for rank, (ip, _) in enumerate(actors):
+            actor_ips[ip].append(rank)
 
-            for rank, ip in enumerate(
-                ray.get([actor.get_ip.remote() for actor in actors])
-            ):
-                actor_ips[ip].append(rank)
-
-            # Get distribution of parts between nodes ({ip:[(part, position),..],..})
-            init_parts_distribution = defaultdict(list)
-            for idx, (ip, part_ref) in enumerate(
-                zip(ray.get(list(parts_ips_ref)), parts_ref)
-            ):
-                init_parts_distribution[ip].append((part_ref, idx))
-
-            num_parts = len(parts_ref)
-            min_parts_per_actor = math.floor(num_parts / num_actors)
-            max_parts_per_actor = math.ceil(num_parts / num_actors)
-            num_actors_with_max_parts = num_parts % num_actors
-
-            row_partitions_by_actors = defaultdict(list)
-            # Fill actors without movement parts between ips
-            for actor_ip, ranks in actor_ips.items():
-                # Loop across actors which are placed on actor_ip
-                for rank in ranks:
-                    num_parts_on_ip = len(init_parts_distribution[actor_ip])
-
-                    # Check that have something to distribute on this ip
-                    if num_parts_on_ip == 0:
-                        break
-                    # Check that node with `actor_ip` has enough parts for minimal
-                    # filling actor with `rank`
-                    if num_parts_on_ip >= min_parts_per_actor:
-                        # Check that node has enough parts for max filling
-                        # actor with `rank`
-                        if (
-                            num_parts_on_ip >= max_parts_per_actor
-                            and num_actors_with_max_parts > 0
-                        ):
-                            pop_slice = slice(0, max_parts_per_actor)
-                            num_actors_with_max_parts -= 1
-                        else:
-                            pop_slice = slice(0, min_parts_per_actor)
-
-                        row_partitions_by_actors[rank].extend(
-                            init_parts_distribution[actor_ip][pop_slice]
-                        )
-                        # Delete parts which we already assign
-                        del init_parts_distribution[actor_ip][pop_slice]
-                    else:
-                        row_partitions_by_actors[rank].extend(
-                            init_parts_distribution[actor_ip]
-                        )
-                        init_parts_distribution[actor_ip] = []
-
-            # Remove empty IPs
-            for ip in list(init_parts_distribution):
-                if len(init_parts_distribution[ip]) == 0:
-                    init_parts_distribution.pop(ip)
-
-            # IP's aren't necessary now
-            init_parts_distribution = [
-                pair for pairs in init_parts_distribution.values() for pair in pairs
-            ]
-
-            # Fill the actors with extra parts (movements data between nodes)
-            for rank in range(len(actors)):
-                num_parts_on_rank = len(row_partitions_by_actors[rank])
-
-                if num_parts_on_rank == max_parts_per_actor or (
-                    num_parts_on_rank == min_parts_per_actor
-                    and num_actors_with_max_parts == 0
-                ):
-                    continue
-
-                if num_actors_with_max_parts > 0:
-                    pop_slice = slice(0, max_parts_per_actor - num_parts_on_rank)
-                    num_actors_with_max_parts -= 1
-                else:
-                    pop_slice = slice(0, min_parts_per_actor - num_parts_on_rank)
-
-                row_partitions_by_actors[rank].extend(
-                    init_parts_distribution[pop_slice]
-                )
-                del init_parts_distribution[pop_slice]
-
-            if len(init_parts_distribution) != 0:
-                raise RuntimeError(
-                    f"Not all partitions were ditributed between actors: {len(init_parts_distribution)} left."
-                )
-
-            row_parts_by_ranks = dict()
-            for rank, pairs_part_pos in dict(row_partitions_by_actors).items():
-                parts, order = zip(*pairs_part_pos)
-                row_parts_by_ranks[rank] = (list(parts), list(order))
-        else:
-            row_parts_by_ranks = {rank: ([], []) for rank in range(len(actors))}
-
-            for rank, (_, order_of_indexes) in data_for_aligning.items():
-                row_parts_by_ranks[rank][1].extend(order_of_indexes)
-                for row_idx in order_of_indexes:
-                    row_parts_by_ranks[rank][0].append(row_partitions[row_idx])
-    else:
-        row_parts_by_ranks = defaultdict(list)
-        _, parts_ref = zip(*row_partitions)
+        # Get distribution of parts between nodes ({ip:[(part, position),..],..})
+        init_parts_distribution = defaultdict(list)
+        for idx, (ip, part_ref) in enumerate(
+            zip(ray.get(list(parts_ips_ref)), parts_ref)
+        ):
+            init_parts_distribution[ip].append((part_ref, idx))
 
         num_parts = len(parts_ref)
         min_parts_per_actor = math.floor(num_parts / num_actors)
         max_parts_per_actor = math.ceil(num_parts / num_actors)
         num_actors_with_max_parts = num_parts % num_actors
 
-        start_idx = 0
-        for rank, actor in enumerate(actors):
+        row_partitions_by_actors = defaultdict(list)
+        # Fill actors without movement parts between ips
+        for actor_ip, ranks in actor_ips.items():
+            # Loop across actors which are placed on actor_ip
+            for rank in ranks:
+                num_parts_on_ip = len(init_parts_distribution[actor_ip])
+
+                # Check that have something to distribute on this ip
+                if num_parts_on_ip == 0:
+                    break
+                # Check that node with `actor_ip` has enough parts for minimal
+                # filling actor with `rank`
+                if num_parts_on_ip >= min_parts_per_actor:
+                    # Check that node has enough parts for max filling
+                    # actor with `rank`
+                    if (
+                        num_parts_on_ip >= max_parts_per_actor
+                        and num_actors_with_max_parts > 0
+                    ):
+                        pop_slice = slice(0, max_parts_per_actor)
+                        num_actors_with_max_parts -= 1
+                    else:
+                        pop_slice = slice(0, min_parts_per_actor)
+
+                    row_partitions_by_actors[rank].extend(
+                        init_parts_distribution[actor_ip][pop_slice]
+                    )
+                    # Delete parts which we already assign
+                    del init_parts_distribution[actor_ip][pop_slice]
+                else:
+                    row_partitions_by_actors[rank].extend(
+                        init_parts_distribution[actor_ip]
+                    )
+                    init_parts_distribution[actor_ip] = []
+
+        # Remove empty IPs
+        for ip in list(init_parts_distribution):
+            if len(init_parts_distribution[ip]) == 0:
+                init_parts_distribution.pop(ip)
+
+        # IP's aren't necessary now
+        init_parts_distribution = [
+            pair for pairs in init_parts_distribution.values() for pair in pairs
+        ]
+
+        # Fill the actors with extra parts (movements data between nodes)
+        for rank in range(len(actors)):
+            num_parts_on_rank = len(row_partitions_by_actors[rank])
+
+            if num_parts_on_rank == max_parts_per_actor or (
+                num_parts_on_rank == min_parts_per_actor
+                and num_actors_with_max_parts == 0
+            ):
+                continue
+
             if num_actors_with_max_parts > 0:
-                num_actor_parts = max_parts_per_actor
+                pop_slice = slice(0, max_parts_per_actor - num_parts_on_rank)
                 num_actors_with_max_parts -= 1
             else:
-                num_actor_parts = min_parts_per_actor
+                pop_slice = slice(0, min_parts_per_actor - num_parts_on_rank)
 
-            idx_slice = slice(start_idx, start_idx + num_actor_parts)
-            row_parts_by_ranks[rank].extend(parts_ref[idx_slice])
-            start_idx += num_actor_parts
+            row_partitions_by_actors[rank].extend(init_parts_distribution[pop_slice])
+            del init_parts_distribution[pop_slice]
+
+        if len(init_parts_distribution) != 0:
+            raise RuntimeError(
+                f"Not all partitions were ditributed between actors: {len(init_parts_distribution)} left."
+            )
+
+        row_parts_by_ranks = dict()
+        for rank, pairs_part_pos in dict(row_partitions_by_actors).items():
+            parts, order = zip(*pairs_part_pos)
+            row_parts_by_ranks[rank] = (list(parts), list(order))
+    else:
+        row_parts_by_ranks = {rank: ([], []) for rank in range(len(actors))}
+
+        for rank, (_, order_of_indexes) in data_for_aligning.items():
+            row_parts_by_ranks[rank][1].extend(order_of_indexes)
+            for row_idx in order_of_indexes:
+                row_parts_by_ranks[rank][0].append(row_partitions[row_idx])
 
     return row_parts_by_ranks
 
 
 def _train(
     dtrain,
-    num_actors,
     params: Dict,
     *args,
+    num_actors=None,
     evals=(),
     **kwargs,
 ):
@@ -590,13 +468,13 @@ def _train(
     ----------
     dtrain : modin.experimental.DMatrix
         Data to be trained against.
-    num_actors : int, optional
-        Number of actors for training. If unspecified, this value will be
-        computed automatically.
     params : dict
         Booster params.
     *args : iterable
         Other parameters for `xgboost.train`.
+    num_actors : int, optional
+        Number of actors for training. If unspecified, this value will be
+        computed automatically.
     evals : list of pairs (modin.experimental.xgboost.DMatrix, str), default: empty
         List of validation sets for which metrics will be evaluated during training.
         Validation metrics will help us track the performance of the model.
@@ -616,14 +494,25 @@ def _train(
 
     assert len(X_row_parts) == len(y_row_parts), "Unaligned train data"
 
-    num_actors = _get_num_actors(
-        num_actors if isinstance(num_actors, int) else "default_train"
-    )
+    num_actors = _get_num_actors(num_actors)
 
     if num_actors > len(X_row_parts):
         num_actors = len(X_row_parts)
 
-    actors, pg = create_actors(num_actors)
+    if evals:
+        min_num_parts = num_actors
+        for (eval_X, _), eval_method in evals:
+            if len(eval_X) < min_num_parts:
+                min_num_parts = len(eval_X)
+                method_name = eval_method
+
+        if num_actors != min_num_parts:
+            num_actors = min_num_parts
+            warnings.warn(
+                f"`num_actors` is set to {num_actors}, because `evals` data with name `{method_name}` has only {num_actors} partition(s)."
+            )
+
+    actors = create_actors(num_actors)
 
     add_as_eval_method = None
     if evals:
@@ -640,7 +529,7 @@ def _train(
                     *X_y, eval_method=eval_method
                 ),
                 eval_X,
-                y_parts=eval_y,
+                eval_y,
             )
 
     # Split data across workers
@@ -650,7 +539,7 @@ def _train(
             *X_y, add_as_eval_method=add_as_eval_method
         ),
         X_row_parts,
-        y_parts=y_row_parts,
+        y_row_parts,
     )
     LOGGER.info(f"Data preparation time: {time.time() - s} s")
 
@@ -660,28 +549,56 @@ def _train(
 
         # Train
         fut = [
-            actor.train.remote(rabit_args, params, *args, **kwargs) for actor in actors
+            actor.train.remote(rabit_args, params, *args, **kwargs)
+            for _, actor in actors
         ]
         # All results should be the same because of Rabit tracking. So we just
         # return the first one.
         result = ray.get(fut[0])
-        remove_placement_group(pg)
         LOGGER.info(f"Training time: {time.time() - s} s")
         return result
 
 
+@ray.remote
+def _map_predict(booster, part, columns, **kwargs):
+    """
+    Run prediction on a remote worker.
+
+    Parameters
+    ----------
+    booster : xgboost.Booster or ray.ObjectRef
+        A trained booster.
+    part : pandas.DataFrame or ray.ObjectRef
+        Partition of full data used for local prediction.
+    columns : list or ray.ObjectRef
+        Columns for the result.
+    **kwargs : dict
+        Other parameters are the same as for ``xgboost.Booster.predict``.
+
+    Returns
+    -------
+    ray.ObjectRef
+        ``ray.ObjectRef`` with partial prediction.
+    """
+    dmatrix = xgb.DMatrix(part)
+    prediction = pandas.DataFrame(
+        booster.predict(dmatrix, **kwargs),
+        index=part.index,
+        columns=columns,
+    )
+    return prediction
+
+
 def _predict(
     booster,
     data,
-    num_actors,
     **kwargs,
 ):
     """
     Run distributed prediction with a trained booster on Ray backend.
 
-    During work it evenly distributes `data` between workers,
-    runs xgb.predict on each worker for subset of `data` and creates
-    Modin DataFrame with prediction results.
+    During execution it runs ``xgb.predict`` on each worker for subset of `data`
+    and creates Modin DataFrame with prediction results.
 
     Parameters
     ----------
@@ -689,11 +606,8 @@ def _predict(
         A trained booster.
     data : modin.experimental.xgboost.DMatrix
         Input data used for prediction.
-    num_actors : int, optional
-        Number of actors for prediction. If unspecified, this value will be
-        computed automatically.
-    **kwargs : dist
-        Other parameters are the same as `xgboost.Booster.predict`.
+    **kwargs : dict
+        Other parameters are the same as for ``xgboost.Booster.predict``.
 
     Returns
     -------
@@ -702,40 +616,39 @@ def _predict(
     """
     s = time.time()
 
-    X_row_parts, _ = data
+    # Get metadata from DMatrix
+    input_index, input_columns, row_lengths = data.metadata
 
-    num_actors = _get_num_actors(
-        num_actors if isinstance(num_actors, int) else "default_predict"
-    )
+    # Infer columns of result
+    def _get_num_columns(booster, n_features, **kwargs):
+        rng = np.random.RandomState(777)
+        test_data = rng.randn(1, n_features)
+        test_predictions = booster.predict(
+            xgb.DMatrix(test_data), validate_features=False, **kwargs
+        )
+        num_columns = (
+            test_predictions.shape[1] if len(test_predictions.shape) > 1 else 1
+        )
+        return num_columns
 
-    if num_actors > len(X_row_parts):
-        num_actors = len(X_row_parts)
-
-    # Create remote actors
-    actors, pg = create_actors(num_actors)
-
-    # Split data across workers
-    _split_data_across_actors(
-        actors,
-        lambda actor, *X: actor.set_predict_data.remote(*X),
-        X_row_parts,
-        is_predict=True,
-    )
-
-    LOGGER.info(f"Data preparation time: {time.time() - s} s")
-    s = time.time()
+    result_num_columns = _get_num_columns(booster, len(input_columns), **kwargs)
+    new_columns = list(range(result_num_columns))
 
+    # Put common data in object store
     booster = ray.put(booster)
+    new_columns_ref = ray.put(new_columns)
 
-    predictions = [
-        tuple(actor.predict.options(num_returns=2).remote(booster, **kwargs))
-        for actor in actors
+    prediction_refs = [
+        _map_predict.remote(booster, part, new_columns_ref, **kwargs)
+        for _, part in data.data
     ]
-
-    ray.wait([part for _, part in predictions], num_returns=len(predictions))
-    remove_placement_group(pg)
-
-    result = from_partitions(predictions, 0)
+    predictions = from_partitions(
+        prediction_refs,
+        0,
+        index=input_index,
+        columns=new_columns,
+        row_lengths=row_lengths,
+        column_widths=[len(new_columns)],
+    )
     LOGGER.info(f"Prediction time: {time.time() - s} s")
-
-    return result
+    return predictions
